@@ -1,24 +1,25 @@
-local configmanager = require "core.configmanager";
-local moduleapi = require "core.moduleapi";
-local show_usage = require "util.prosodyctl".show_usage;
-local show_warning = require "util.prosodyctl".show_warning;
-local is_prosody_running = require "util.prosodyctl".isrunning;
-local parse_args = require "util.argparse".parse;
-local dependencies = require "util.dependencies";
+local configmanager = require "prosody.core.configmanager";
+local moduleapi = require "prosody.core.moduleapi";
+local show_usage = require "prosody.util.prosodyctl".show_usage;
+local show_warning = require "prosody.util.prosodyctl".show_warning;
+local is_prosody_running = require "prosody.util.prosodyctl".isrunning;
+local parse_args = require "prosody.util.argparse".parse;
+local dependencies = require "prosody.util.dependencies";
 local socket = require "socket";
 local socket_url = require "socket.url";
-local jid_split = require "util.jid".prepped_split;
-local modulemanager = require "core.modulemanager";
-local async = require "util.async";
-local httputil = require "util.http";
+local jid_split = require "prosody.util.jid".prepped_split;
+local modulemanager = require "prosody.core.modulemanager";
+local async = require "prosody.util.async";
+local httputil = require "prosody.util.http";
+local human_units = require "prosody.util.human.units";
 
 local function api(host)
 	return setmetatable({ name = "prosodyctl.check"; host = host; log = prosody.log }, { __index = moduleapi })
 end
 
 local function check_ojn(check_type, target_host)
-	local http = require "net.http"; -- .new({});
-	local json = require "util.json";
+	local http = require "prosody.net.http"; -- .new({});
+	local json = require "prosody.util.json";
 
 	local response, err = async.wait_for(http.request(
 		("https://observe.jabber.network/api/v1/check/%s"):format(httputil.urlencode(check_type)),
@@ -46,7 +47,7 @@ local function check_ojn(check_type, target_host)
 end
 
 local function check_probe(base_url, probe_module, target)
-	local http = require "net.http"; -- .new({});
+	local http = require "prosody.net.http"; -- .new({});
 	local params = httputil.formencode({ module = probe_module; target = target })
 	local response, err = async.wait_for(http.request(base_url .. "?" .. params));
 
@@ -67,8 +68,8 @@ local function check_probe(base_url, probe_module, target)
 end
 
 local function check_turn_service(turn_service, ping_service)
-	local ip = require "util.ip";
-	local stun = require "net.stun";
+	local ip = require "prosody.util.ip";
+	local stun = require "prosody.net.stun";
 
 	local result = { warnings = {} };
 
@@ -170,7 +171,7 @@ local function check_turn_service(turn_service, ping_service)
 		result.error = "TURN server did not response to allocation request: "..err;
 		return result;
 	elseif alloc_response:is_err_resp() then
-		result.error = ("TURN allocation failed: %d (%s)"):format(alloc_response:get_error());
+		result.error = ("TURN server failed to create allocation: %d (%s)"):format(alloc_response:get_error());
 		return result;
 	elseif not alloc_response:is_success_resp() then
 		result.error = ("Unexpected TURN response: %d (%s)"):format(alloc_response:get_type());
@@ -319,18 +320,20 @@ local function check(arg)
 		print("Error: Unknown parameter: "..opts_info);
 		return 1;
 	end
-	local array = require "util.array";
-	local set = require "util.set";
-	local it = require "util.iterators";
+	local array = require "prosody.util.array";
+	local set = require "prosody.util.set";
+	local it = require "prosody.util.iterators";
 	local ok = true;
+	local function contains_match(hayset, needle) for member in hayset do if member:find(needle) then return true end end end
 	local function disabled_hosts(host, conf) return host ~= "*" and conf.enabled ~= false; end
-	local function enabled_hosts() return it.filter(disabled_hosts, pairs(configmanager.getconfig())); end
-	if not (what == nil or what == "disabled" or what == "config" or what == "dns" or what == "certs" or what == "connectivity" or what == "turn") then
-		show_warning("Don't know how to check '%s'. Try one of 'config', 'dns', 'certs', 'disabled', 'turn' or 'connectivity'.", what);
-		show_warning("Note: The connectivity check will connect to a remote server.");
-		return 1;
-	end
-	if not what or what == "disabled" then
+	local function is_user_host(host, conf) return host ~= "*" and conf.component_module == nil; end
+	local function is_component_host(host, conf) return host ~= "*" and conf.component_module ~= nil; end
+	local function enabled_hosts() return it.filter(disabled_hosts, it.sorted_pairs(configmanager.getconfig())); end
+	local function enabled_user_hosts() return it.filter(is_user_host, it.sorted_pairs(configmanager.getconfig())); end
+	local function enabled_components() return it.filter(is_component_host, it.sorted_pairs(configmanager.getconfig())); end
+
+	local checks = {};
+	function checks.disabled()
 		local disabled_hosts_set = set.new();
 		for host in it.filter("*", pairs(configmanager.getconfig())) do
 			if api(host):get_option_boolean("enabled") == false then
@@ -345,7 +348,7 @@ local function check(arg)
 			print""
 		end
 	end
-	if not what or what == "config" then
+	function checks.config()
 		print("Checking config...");
 
 		if what == "config" then
@@ -520,9 +523,9 @@ local function check(arg)
 			end
 			for k, v in pairs(modules) do
 				if type(k) ~= "number" or type(v) ~= "string" then
-					print("    The " .. name .. " in the " .. host .. " section should not be a map of " .. type(k) .. " to " .. type(v)
-									.. " but a list of strings, e.g.");
+					print("    The " .. name .. " in the " .. host .. " section should be a list of strings, e.g.");
 					print("    " .. name .. " = { \"name_of_module\", \"another_plugin\", }")
+					print("    It should not contain key = value pairs, try putting them outside the {} brackets.");
 					ok = false
 					break
 				end
@@ -635,6 +638,12 @@ local function check(arg)
 			print("    Both mod_pep_simple and mod_pep are enabled but they conflict");
 			print("    with each other. Remove one.");
 		end
+		if all_modules:contains("posix") then
+			print("");
+			print("    mod_posix is loaded in your configuration file, but it has");
+			print("    been deprecated. You can safely remove it.");
+		end
+
 		for host, host_config in pairs(config) do --luacheck: ignore 213/host
 			if type(rawget(host_config, "storage")) == "string" and rawget(host_config, "default_storage") then
 				print("");
@@ -642,6 +651,15 @@ local function check(arg)
 				break;
 			end
 		end
+
+		for host, host_config in pairs(config) do --luacheck: ignore 213/host
+			if type(rawget(host_config, "storage")) == "string" and rawget(host_config, "default_storage") then
+				print("");
+				print("    The 'default_storage' option is not needed if 'storage' is set to a string.");
+				break;
+			end
+		end
+
 		local require_encryption = set.intersection(all_options, set.new({
 			"require_encryption", "c2s_require_encryption", "s2s_require_encryption"
 		})):empty();
@@ -716,12 +734,17 @@ local function check(arg)
 			local orphan_components = {};
 			local referenced_components = set.new();
 			local enabled_hosts_set = set.new();
+			local invalid_disco_items = {};
 			for host in it.filter("*", pairs(configmanager.getconfig())) do
 				local hostapi = api(host);
 				if hostapi:get_option_boolean("enabled", true) then
 					enabled_hosts_set:add(host);
 					for _, disco_item in ipairs(hostapi:get_option_array("disco_items", {})) do
-						referenced_components:add(disco_item[1]);
+						if type(disco_item[1]) == "string" then
+							referenced_components:add(disco_item[1]);
+						else
+							invalid_disco_items[host] = true;
+						end
 					end
 				end
 			end
@@ -735,6 +758,18 @@ local function check(arg)
 					end
 				end
 			end
+
+			if next(invalid_disco_items) ~= nil then
+				print("");
+				print("    Some hosts in your configuration file have an invalid 'disco_items' option.");
+				print("    This may cause further errors, such as unreferenced components.");
+				print("");
+				for host in it.sorted_pairs(invalid_disco_items) do
+					print("      - "..host);
+				end
+				print("");
+			end
+
 			if #orphan_components > 0 then
 				table.sort(orphan_components);
 				print("");
@@ -750,7 +785,7 @@ local function check(arg)
 
 		-- Check hostname validity
 		do
-			local idna = require "util.encodings".idna;
+			local idna = require "prosody.util.encodings".idna;
 			local invalid_hosts = {};
 			local alabel_hosts = {};
 			for host in it.filter("*", pairs(configmanager.getconfig())) do
@@ -793,22 +828,38 @@ local function check(arg)
 
 			if #invalid_hosts > 0 or #alabel_hosts > 0 then
 				print("");
-				print("WARNING: Changing the name of a VirtualHost in Prosody's config file");
-				print("         WILL NOT migrate any existing data (user accounts, etc.) to the new name.");
+				print("    WARNING: Changing the name of a VirtualHost in Prosody's config file");
+				print("             WILL NOT migrate any existing data (user accounts, etc.) to the new name.");
 				ok = false;
 			end
 		end
 
+		-- Check features
+		do
+			local missing_features = {};
+			for host in enabled_user_hosts() do
+				local all_features = checks.features(host, true);
+				if not all_features then
+					table.insert(missing_features, host);
+				end
+			end
+			if #missing_features > 0 then
+				print("");
+				print("    Some of your hosts may be missing features due to a lack of configuration.");
+				print("    For more details, use the 'prosodyctl check features' command.");
+			end
+		end
+
 		print("Done.\n");
 	end
-	if not what or what == "dns" then
-		local dns = require "net.dns";
+	function checks.dns()
+		local dns = require "prosody.net.dns";
 		pcall(function ()
-			local unbound = require"net.unbound";
+			local unbound = require"prosody.net.unbound";
 			dns = unbound.dns;
 		end)
-		local idna = require "util.encodings".idna;
-		local ip = require "util.ip";
+		local idna = require "prosody.util.encodings".idna;
+		local ip = require "prosody.util.ip";
 		local global = api("*");
 		local c2s_ports = global:get_option_set("c2s_ports", {5222});
 		local s2s_ports = global:get_option_set("s2s_ports", {5269});
@@ -871,7 +922,7 @@ local function check(arg)
 			end
 		end
 
-		local local_addresses = require"util.net".local_addresses() or {};
+		local local_addresses = require"prosody.util.net".local_addresses() or {};
 
 		for addr in it.values(local_addresses) do
 			if not ip.new_ip(addr).private then
@@ -904,7 +955,11 @@ local function check(arg)
 
 		local unknown_addresses = set.new();
 
-		for jid in enabled_hosts() do
+		local function is_valid_domain(domain)
+			return idna.to_ascii(domain) ~= nil;
+		end
+
+		for jid in it.filter(is_valid_domain, enabled_hosts()) do
 			local all_targets_ok, some_targets_ok = true, false;
 			local node, host = jid_split(jid);
 
@@ -1038,9 +1093,6 @@ local function check(arg)
 			end
 
 			local known_http_modules = set.new { "bosh"; "http_files"; "http_file_share"; "http_openmetrics"; "websocket" };
-			local function contains_match(hayset, needle)
-				for member in hayset do if member:find(needle) then return true end end
-			end
 
 			if modules:contains("http") or not set.intersection(modules, known_http_modules):empty()
 				or contains_match(modules, "^http_") or contains_match(modules, "_web$") then
@@ -1176,11 +1228,14 @@ local function check(arg)
 			ok = false;
 		end
 	end
-	if not what or what == "certs" then
+	function checks.certs()
 		local cert_ok;
 		print"Checking certificates..."
-		local x509_verify_identity = require"util.x509".verify_identity;
-		local create_context = require "core.certmanager".create_context;
+		local x509_verify_identity = require"prosody.util.x509".verify_identity;
+		local use_dane = configmanager.get("*", "use_dane");
+		local pem2der = require"prosody.util.x509".pem2der;
+		local sha256 = require"prosody.util.hashes".sha256;
+		local create_context = require "prosody.core.certmanager".create_context;
 		local ssl = dependencies.softreq"ssl";
 		-- local datetime_parse = require"util.datetime".parse_x509;
 		local load_cert = ssl and ssl.loadcertificate;
@@ -1193,13 +1248,14 @@ local function check(arg)
 			cert_ok = false
 		else
 			for host in it.filter(skip_bare_jid_hosts, enabled_hosts()) do
+				local modules = modulemanager.get_modules_for_host(host);
 				print("Checking certificate for "..host);
 				-- First, let's find out what certificate this host uses.
 				local host_ssl_config = configmanager.rawget(host, "ssl")
 					or configmanager.rawget(host:match("%.(.*)"), "ssl");
 				local global_ssl_config = configmanager.rawget("*", "ssl");
-				local ok, err, ssl_config = create_context(host, "server", host_ssl_config, global_ssl_config);
-				if not ok then
+				local ctx_ok, err, ssl_config = create_context(host, "server", host_ssl_config, global_ssl_config);
+				if not ctx_ok then
 					print("  Error: "..err);
 					cert_ok = false
 				elseif not ssl_config.certificate then
@@ -1234,17 +1290,39 @@ local function check(arg)
 						elseif not cert:validat(os.time() + 86400*31) then
 							print("    Certificate expires within one month.")
 						end
-						if select(2, modulemanager.get_modules_for_host(host)) == nil
-							and not x509_verify_identity(host, "_xmpp-client", cert) then
+						if modules:contains("c2s") and not x509_verify_identity(host, "_xmpp-client", cert) then
 							print("    Not valid for client connections to "..host..".")
 							cert_ok = false
 						end
-						if (not (api(host):get_option_boolean("anonymous_login", false)
-							or api(host):get_option_string("authentication", "internal_hashed") == "anonymous"))
-							and not x509_verify_identity(host, "_xmpp-server", cert) then
+						local anon = api(host):get_option_string("authentication", "internal_hashed") == "anonymous";
+						local anon_s2s = api(host):get_option_boolean("allow_anonymous_s2s", false);
+						if modules:contains("s2s") and (anon_s2s or not anon) and not x509_verify_identity(host, "_xmpp-server", cert) then
 							print("    Not valid for server-to-server connections to "..host..".")
 							cert_ok = false
 						end
+
+						local known_http_modules = set.new { "bosh"; "http_files"; "http_file_share"; "http_openmetrics"; "websocket" };
+						local http_loaded = modules:contains("http")
+							or not set.intersection(modules, known_http_modules):empty()
+							or contains_match(modules, "^http_")
+							or contains_match(modules, "_web$");
+
+						local http_host = api(host):get_option_string("http_host", host);
+						if api(host):get_option_string("http_external_url") then
+							-- Assumed behind a reverse proxy
+							http_loaded = false;
+						end
+						if http_loaded and not x509_verify_identity(http_host, nil, cert) then
+							print("    Not valid for HTTPS connections to "..http_host..".")
+							cert_ok = false
+						end
+						if use_dane then
+							if cert.pubkey then
+								print("    DANE: TLSA 3 1 1 "..sha256(pem2der(cert:pubkey()), true))
+							elseif cert.pem then
+								print("    DANE: TLSA 3 0 1 "..sha256(pem2der(cert:pem()), true))
+							end
+						end
 					end
 				end
 			end
@@ -1257,7 +1335,7 @@ local function check(arg)
 		print("")
 	end
 	-- intentionally not doing this by default
-	if what == "connectivity" then
+	function checks.connectivity()
 		local _, prosody_is_running = is_prosody_running();
 		if api("*"):get_option_string("pidfile") and not prosody_is_running then
 			print("Prosody does not appear to be running, which is required for this test.");
@@ -1349,7 +1427,7 @@ local function check(arg)
 		print("Note: It does not ensure that the check actually reaches this specific prosody instance.")
 	end
 
-	if not what or what == "turn" then
+	function checks.turn()
 		local turn_enabled_hosts = {};
 		local turn_services = {};
 
@@ -1425,6 +1503,333 @@ local function check(arg)
 		end
 	end
 
+	function checks.features(check_host, quiet)
+		if not quiet then
+			print("Feature report");
+		end
+
+		local common_subdomains = {
+			http_file_share = "share";
+			muc = "groups";
+		};
+
+		local recommended_component_modules = {
+			muc = { "muc_mam" };
+		};
+
+		local function print_feature_status(feature, host)
+			if quiet then return; end
+			print("", feature.ok and "OK" or "(!)", feature.name);
+			if feature.desc then
+				print("", "", feature.desc);
+				print("");
+			end
+			if not feature.ok then
+				if feature.lacking_modules then
+					table.sort(feature.lacking_modules);
+					print("", "", "Suggested modules: ");
+					for _, module in ipairs(feature.lacking_modules) do
+						print("", "", ("  - %s: https://prosody.im/doc/modules/mod_%s"):format(module, module));
+					end
+				end
+				if feature.lacking_components then
+					table.sort(feature.lacking_components);
+					for _, component_module in ipairs(feature.lacking_components) do
+						local subdomain = common_subdomains[component_module];
+						local recommended_mods = recommended_component_modules[component_module];
+						if subdomain then
+							print("", "", "Suggested component:");
+							print("");
+							print("", "", "", ("-- Documentation: https://prosody.im/doc/modules/mod_%s"):format(component_module));
+							print("", "", "", ("Component %q %q"):format(subdomain.."."..host, component_module));
+							if recommended_mods then
+								print("", "", "", "    modules_enabled = {");
+								table.sort(recommended_mods);
+								for _, mod in ipairs(recommended_mods) do
+									print("", "", "", ("        %q;"):format(mod));
+								end
+								print("", "", "", "    }");
+							end
+						else
+							print("", "", ("Suggested component: %s"):format(component_module));
+						end
+					end
+					print("");
+					print("", "", "If you have already configured any of these components, they may not be");
+					print("", "", "linked correctly to "..host..". For more info see https://prosody.im/doc/components");
+				end
+				if feature.lacking_component_modules then
+					table.sort(feature.lacking_component_modules, function (a, b)
+						return a.host < b.host;
+					end);
+					for _, problem in ipairs(feature.lacking_component_modules) do
+						local hostapi = api(problem.host);
+						local current_modules_enabled = hostapi:get_option_array("modules_enabled", {});
+						print("", "", ("Component %q is missing the following modules: %s"):format(problem.host, table.concat(problem.missing_mods)));
+						print("");
+						print("","", "Add the missing modules to your modules_enabled under the Component, like this:");
+						print("");
+						print("");
+						print("", "", "", ("-- Documentation: https://prosody.im/doc/modules/mod_%s"):format(problem.component_module));
+						print("", "", "", ("Component %q %q"):format(problem.host, problem.component_module));
+						print("", "", "", ("    modules_enabled = {"));
+						for _, mod in ipairs(current_modules_enabled) do
+							print("", "", "", ("        %q;"):format(mod));
+						end
+						for _, mod in ipairs(problem.missing_mods) do
+							print("", "", "", ("        %q; -- Add this!"):format(mod));
+						end
+						print("", "", "", ("    }"));
+					end
+				end
+			end
+			if feature.meta then
+				for k, v in it.sorted_pairs(feature.meta) do
+					print("", "", (" - %s: %s"):format(k, v));
+				end
+			end
+			print("");
+		end
+
+		local all_ok = true;
+
+		local config = configmanager.getconfig();
+
+		local f, s, v;
+		if check_host then
+			f, s, v = it.values({ check_host });
+		else
+			f, s, v = enabled_user_hosts();
+		end
+
+		for host in f, s, v do
+			local modules_enabled = set.new(config["*"].modules_enabled);
+			modules_enabled:include(set.new(config[host].modules_enabled));
+
+			-- { [component_module] = { hostname1, hostname2, ... } }
+			local host_components = setmetatable({}, { __index = function (t, k) return rawset(t, k, {})[k]; end });
+
+			do
+				local hostapi = api(host);
+
+				-- Find implicitly linked components
+				for other_host in enabled_components() do
+					local parent_host = other_host:match("^[^.]+%.(.+)$");
+					if parent_host == host then
+						local component_module = configmanager.get(other_host, "component_module");
+						if component_module then
+							table.insert(host_components[component_module], other_host);
+						end
+					end
+				end
+
+				-- And components linked explicitly
+				for _, disco_item in ipairs(hostapi:get_option_array("disco_items", {})) do
+					local other_host = disco_item[1];
+					if type(other_host) == "string" then
+						local component_module = configmanager.get(other_host, "component_module");
+						if component_module then
+							table.insert(host_components[component_module], other_host);
+						end
+					end
+				end
+			end
+
+			local current_feature;
+
+			local function check_module(suggested, alternate, ...)
+				if set.intersection(modules_enabled, set.new({suggested, alternate, ...})):empty() then
+					current_feature.lacking_modules = current_feature.lacking_modules or {};
+					table.insert(current_feature.lacking_modules, suggested);
+				end
+			end
+
+			local function check_component(suggested, alternate, ...)
+				local found;
+				for _, component_module in ipairs({ suggested, alternate, ... }) do
+					found = host_components[component_module][1];
+					if found then
+						local enabled_component_modules = api(found):get_option_inherited_set("modules_enabled");
+						local recommended_mods = recommended_component_modules[component_module];
+						if recommended_mods then
+							local missing_mods = {};
+							for _, mod in ipairs(recommended_mods) do
+								if not enabled_component_modules:contains(mod) then
+									table.insert(missing_mods, mod);
+								end
+							end
+							if #missing_mods > 0 then
+								if not current_feature.lacking_component_modules then
+									current_feature.lacking_component_modules = {};
+								end
+								table.insert(current_feature.lacking_component_modules, {
+									host = found;
+									component_module = component_module;
+									missing_mods = missing_mods;
+								});
+							end
+						end
+						break;
+					end
+				end
+				if not found then
+					current_feature.lacking_components = current_feature.lacking_components or {};
+					table.insert(current_feature.lacking_components, suggested);
+				end
+				return found;
+			end
+
+			local features = {
+				{
+					name = "Basic functionality";
+					desc = "Support for secure connections, authentication and messaging";
+					check = function ()
+						check_module("disco");
+						check_module("roster");
+						check_module("saslauth");
+						check_module("tls");
+					end;
+				};
+				{
+					name = "Multi-device messaging and data synchronization";
+					desc = "Multiple clients connected to the same account stay in sync";
+					check = function ()
+						check_module("carbons");
+						check_module("mam");
+						check_module("bookmarks");
+						check_module("pep");
+					end;
+				};
+				{
+					name = "Mobile optimizations";
+					desc = "Help mobile clients reduce battery and data usage";
+					check = function ()
+						check_module("smacks");
+						check_module("csi_simple", "csi_battery_saver");
+					end;
+				};
+				{
+					name = "Web connections";
+					desc = "Allow connections from browser-based web clients";
+					check = function ()
+						check_module("bosh");
+						check_module("websocket");
+					end;
+				};
+				{
+					name = "User profiles";
+					desc = "Enable users to publish profile information";
+					check = function ()
+						check_module("vcard_legacy", "vcard");
+					end;
+				};
+				{
+					name = "Blocking";
+					desc = "Block communication with chosen entities";
+					check = function ()
+						check_module("blocklist");
+					end;
+				};
+				{
+					name = "Push notifications";
+					desc = "Receive notifications on platforms that don't support persistent connections";
+					check = function ()
+						check_module("cloud_notify");
+					end;
+				};
+				{
+					name = "Audio/video calls and P2P";
+					desc = "Assist clients in setting up connections between each other";
+					check = function ()
+						check_module(
+							"turn_external",
+							"external_services",
+							"turncredentials",
+							"extdisco"
+						);
+					end;
+				};
+				{
+					name = "File sharing";
+					desc = "Sharing of files to groups and offline users";
+					check = function (self)
+						local service = check_component("http_file_share", "http_upload", "http_upload_external");
+						if service then
+							local size_limit;
+							if api(service):get_option("component_module") == "http_file_share" then
+								size_limit = api(service):get_option_number("http_file_share_size_limit", 10*1024*1024);
+							end
+							if size_limit then
+								self.meta = {
+									["Size limit"] = human_units.format(size_limit, "b", "b");
+								};
+							end
+						end
+					end;
+				};
+				{
+					name = "Group chats";
+					desc = "Create group chats and channels";
+					check = function ()
+						check_component("muc");
+					end;
+				};
+			};
+
+			if not quiet then
+				print(host);
+			end
+
+			for _, feature in ipairs(features) do
+				current_feature = feature;
+				feature:check();
+				feature.ok = (
+					not feature.lacking_modules and
+					not feature.lacking_components and
+					not feature.lacking_component_modules
+				);
+				-- For improved presentation, we group the (ok) and (not ok) features
+				if feature.ok then
+					print_feature_status(feature, host);
+				end
+			end
+
+			for _, feature in ipairs(features) do
+				if not feature.ok then
+					all_ok = false;
+					print_feature_status(feature, host);
+				end
+			end
+
+			if not quiet then
+				print("");
+			end
+		end
+
+		return all_ok;
+	end
+
+	if what == nil or what == "all" then
+		local ret;
+		ret = checks.disabled();
+		if ret ~= nil then return ret; end
+		ret = checks.config();
+		if ret ~= nil then return ret; end
+		ret = checks.dns();
+		if ret ~= nil then return ret; end
+		ret = checks.certs();
+		if ret ~= nil then return ret; end
+		ret = checks.turn();
+		if ret ~= nil then return ret; end
+	elseif checks[what] then
+		local ret = checks[what]();
+		if ret ~= nil then return ret; end
+	else
+		show_warning("Don't know how to check '%s'. Try one of 'config', 'dns', 'certs', 'disabled', 'turn' or 'connectivity'.", what);
+		show_warning("Note: The connectivity check will connect to a remote server.");
+		return 1;
+	end
+
 	if not ok then
 		print("Problems found, see above.");
 	else
